#------------------------------------------------------------------------------------------------
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#------------------------------------------------------------------------------------------------
# THIS PLUGIN DEVELOPED AND TESTED UNDER WINDOWS VISTA ONLY.  MAY OR MAY NOT FUNCTION ON A MAC.
#------------------------------------------------------------------------------------------------
#    Name:	Follow Me Helix
#      By:	sdmitch
#   Usage:	creates a helix around a selected centerline.
#    Note:	On centerlines with sharp bends and short segments results is usally improved by reducing 
#						the linear increment and smoothing by spline fit.  The smoothness of the helix is a 
#						function of the angular increment.  The helix winds clockwise with a positive angular
#						increment and counter-clockwise with a negative angular increment.
#    Date:	Jan 2012
#------------------------------------------------------------------------------------------------
require 'sketchup'
#
if not file_loaded?(File.basename(__FILE__))
    UI.menu('Plugins').add_item('Follow Me Helix') { SDM_FMH.helix_maker }
	file_loaded(File.basename(__FILE__))
end
#
module SDM_FMH

	def self.helix_maker
		sel=Sketchup.active_model.selection
		if sel.empty? then
			UI.messagebox "Select the entities that will be the shape and length of the helix you want to create"
			return
		else
			@radius=1.feet if !@radius;@dist_inc=1.inch if !@dist_inc;@ang_inc=30.0 if !@ang_inc
			prompts=["Radius of helix:","linear Increment:","Angle Increment:"]
			ans=UI.inputbox(prompts,[@radius,@dist_inc,@ang_inc],"Helix Generator")
			if !ans then;return;end
			@radius=ans[0].to_l;@dist_inc = ans[1].to_l; @ang_inc=ans[2].to_f
			verts = sort_edges(sel); last = verts.length - 1; 
			min_length=1e6;max_angle=0.0;curve_length=0.0;last_vector=nil
			if last <=0 then;return;end
			for i in 1..last
				seg_length=verts[i-1].distance(verts[i])
				min_length=[min_length,seg_length].min
				next_vector=verts[i-1].vector_to(verts[i]).normalize
				if last_vector then
					ang_btwn=last_vector.angle_between(next_vector)
					max_angle=[ang_btwn,max_angle].max
				end
				curve_length += seg_length;last_vector=next_vector
			end
			# puts "min_length=#{min_length}\nmax_angle=#{max_angle}"
			@max_inc = (min_length/6).to_l; 
			if @dist_inc > @max_inc
				@dist_inc = @max_inc if(UI.messagebox("Linear increment may be to big, #{@max_inc} or less is recommended.  Reduce?" ,MB_YESNO)==6)
			end
			###
			if max_angle >= 30.degrees
				verts=spline(verts) if(UI.messagebox("Some angles between segments exceed 30 degrees.  Do spline fit to smooth?",MB_YESNO)==6)
			end
			###
			verts=equal_segments(@dist_inc,verts)
			###
			Sketchup.active_model.start_operation("Helix")
			angle = 0; last_vector=nil;pts=[];last = verts.length - 1; puts "last=#{last}"
			next_vector=verts[0].vector_to(verts[1]).normalize
			xaxis,yaxis,zaxis=next_vector.axes
			xaxis.length = @radius; yaxis.length = @radius
			cosa = Math.cos(angle.degrees); sina = Math.sin(angle.degrees)
			vec = Geom::Vector3d.linear_combination(cosa,xaxis,sina,yaxis)
			pts.push verts[0].offset(vec)
			# Sketchup.active_model.entities.add_line(verts[0],pts[-1]);# spoke cl to helix
			1.upto(last) do |i|
				if i < last
					angle += @ang_inc; angle += 360.0 if angle < 0.0; angle %= 360.0
					cosa = Math.cos(angle.degrees); sina = Math.sin(angle.degrees)
					last_vector=next_vector;next_vector=verts[i].vector_to(verts[i+1]).normalize
					vector=Geom::Vector3d.linear_combination(0.5,last_vector,0.5,next_vector)
					xaxis,yaxis,zaxis=vector.axes;xaxis.length = @radius; yaxis.length = @radius
					vec = Geom::Vector3d.linear_combination(cosa,xaxis,sina,yaxis)
					pts.push verts[i].offset(vec)
					# Sketchup.active_model.entities.add_line(verts[i],pts[-1]);# spoke cl to helix
				else
					dist=verts[-2].distance(verts[-1]);pct=dist/@dist_inc
					angle += @ang_inc*pct; angle += 360.0 if angle < 0.0; angle %= 360.0
					cosa = Math.cos(angle.degrees); sina = Math.sin(angle.degrees)
					next_vector=verts[-2].vector_to(verts[-1]).normalize;#puts next_vector
					xaxis,yaxis,zaxis=next_vector.axes;xaxis.length = @radius; yaxis.length = @radius
					vec = Geom::Vector3d.linear_combination(cosa,xaxis,sina,yaxis)
					pts.push verts[-1].offset(vec)
					# Sketchup.active_model.entities.add_line(verts[i],pts[-1]);# spoke cl to helix
				end
			end
			Sketchup.set_status_text "Adding #{pts.length} point helix curve to model.  This might take awhile."
			curve=Sketchup.active_model.entities.add_curve(pts)
			Sketchup.active_model.commit_operation
		end
	end

	def self.sort_edges(sel)
		edges=sel.select {|x| x.is_a? Sketchup::Edge}
	#	---------------- This code copied from FAK.rb ------------------
		return if edges.length==0
		rp=[]	# the vertices
		while edges.length!=0
			start=nil
			edges.each{|x|
			start=x.start; break  if (x.start.edges&edges).length==1
			start=x.end; break if (x.end.edges&edges).length==1
			}
			vts=[]; curve=[]
			while start
				next_edge=(start.edges&edges-curve)[0]
				if next_edge
					curve << next_edge
					vts << start
					start=next_edge.other_vertex(start)
				else
					vts << start
					start=nil
				end
			end
			edges-=curve
			rp += vts 
		end
		rp.collect!{|x| x.position}
		return rp
	end

	def self.spline(pts)
		x = []; y = []; z = []; k = 1; spts = []
		#	Add an extra point at each end
		if pts[0] == pts[-1] then #closed loop
			pts.push pts[1]
			pts.insert(0,pts[-2])
		else
			pt=pts[0].offset(pts[1].vector_to(pts[0]),1)
			pts.insert(0,pt)
			pt=pts[-1].offset(pts[-2].vector_to(pts[-1]),1)
			pts.push pt
		end
		@np = pts.length-1
		#	Get the x,y,z values of the points for spline interpolation
		for i in (0..@np)
			x[i] = pts[i].x
			y[i] = pts[i].y
			z[i] = pts[i].z
		end
		#	Catmull_Rom interpolation used to define the Spline
		spts[0] = pts[1]	# start spline with first point
		for i in (1..@np-2)
			span = pts[i].distance(pts[i+1]);
			if  span >= @max_inc
				numpts = (span / @dist_inc).ceil
				numpts = [numpts,3].max
				numpts = [numpts,10].min; 
				inc = 1.0 / numpts.to_f; 
				inc.step(1.0,inc) do |mu|
					xs = cat_rom(x[i-1],x[i],x[i+1],x[i+2],mu)
					ys = cat_rom(y[i-1],y[i],y[i+1],y[i+2],mu)
					zs = cat_rom(z[i-1],z[i],z[i+1],z[i+2],mu)
					spts[k] = [xs,ys,zs]
					k += 1
				end
			end
		end
		return spts
	end

	def self.cat_rom(p0,p1,p2,p3,mu)
		a0 = -0.5*p0 + 1.5*p1 - 1.5*p2 + 0.5*p3;
		a1 = p0 - 2.5*p1 + 2*p2 - 0.5*p3;
		a2 = -0.5*p0 + 0.5*p2;
		a3 = p1;
		return (a0*mu**3+a1*mu**2+a2*mu+a3)
	end
	
	def self.equal_segments(rad,pts)
		#	compute the end points of the equal segments
		arr = []; arr.push pts[0]
		for n in 1...pts.length
			begin
				pt=self.line_sphere_intersection(pts[n-1],pts[n],arr[-1],rad)
				if pt.length > 0
					if self.is_between(pt[0],pts[n-1],pts[n])
						arr.push  pt[0]
					end
				end
			end until arr[-1] != pt[0]
		end#for
		arr.push pts[-1]
		return arr
	end
	
	def  self.is_between(p0,p1,p2)
		d1=p0.distance p1 
		d2=p0.distance p2 
		d3=p1.distance p2 
		return true if ([d1,d2].max <= d3)
		return false # p0 is not between p1 and p2
	end
	
	def  self.line_sphere_intersection(pt0,pt1,ctr,ros)
		dx = pt1.x - pt0.x
		dy = pt1.y - pt0.y
		dz = pt1.z - pt0.z
		a =  dx**2 + dy**2 + dz**2
		b =  2*( dx*(pt0.x - ctr.x) + dy*(pt0.y - ctr.y) + dz*(pt0.z - ctr.z))
		c =  ctr.x**2 + ctr.y**2 + ctr.z**2 + pt0.x**2 + pt0.y**2 + pt0.z**2 - 2*( ctr.x*pt0.x + ctr.y*pt0.y + ctr.z*pt0.z ) - ros**2
		i =   b * b - 4 * a * c 
		p = [] # initialize point array
		if ( i == 0.0 ) # one intersection
			mu = -b/(2*a) ;
			x = pt0.x + mu*dx;
			y = pt0.y + mu*dy;
			z = pt0.z + mu*dz;
			p.push Geom::Point3d.new(x,y,z)
		elsif ( i > 0.0 ) # two intersections
			# first intersection
			mu = (-b + Math.sqrt( (b**2) - 4*a*c )) / (2*a);
			x = pt0.x + mu*dx;
			y = pt0.y + mu*dy;
			z = pt0.z + mu*dz;
			p.push Geom::Point3d.new(x,y,z)
			# second intersection
			mu = (-b - Math.sqrt((b**2) - 4*a*c )) / (2*a);
			x = pt0.x + mu*dx;
			y = pt0.y + mu*dy;
			z = pt0.z + mu*dz;
			p.push Geom::Point3d.new(x,y,z)
		end
		return p;
	end
	
end
